@@ -1191,89 +1191,438 @@ class TestContextCredentials(unittest.TestCase):
             assert self.ACCESS_KEY in str(request.headers.get('Authorization'))
 
 
-class TestFeatureIdRegistered:
-    @patch(
-        "botocore.utils.InstanceMetadataFetcher.retrieve_iam_role_credentials"
+@pytest.mark.parametrize(
+    "environ_vars,fake_credentials,patches,expected_feature_id",
+    [
+        # Test case 1: IMDS credentials
+        (
+            {},
+            {},
+            [
+                patch(
+                    "botocore.utils.InstanceMetadataFetcher.retrieve_iam_role_credentials",
+                    return_value={
+                        "role_name": "FAKEROLE",
+                        "access_key": "FAKEACCESSKEY",
+                        "secret_key": "FAKESECRET",
+                        "token": "FAKETOKEN",
+                        "expiry_time": "2099-01-01T00:00:00Z",
+                    },
+                ),
+                patch(
+                    "botocore.credentials.ContainerProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.ConfigProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.SharedCredentialProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.EnvProvider.load", return_value=None
+                ),
+            ],
+            '0',
+        ),
+        # Test case 2: HTTP credentials (container)
+        (
+            {
+                'AWS_CONTAINER_CREDENTIALS_FULL_URI': 'http://localhost/foo',
+                'AWS_CONTAINER_AUTHORIZATION_TOKEN': 'Basic auth-token',
+            },
+            {},
+            [
+                patch(
+                    "botocore.credentials.ContainerMetadataFetcher.retrieve_full_uri",
+                    return_value={
+                        "AccessKeyId": "FAKEACCESSKEY",
+                        "SecretAccessKey": "FAKESECRET",
+                        "Token": "FAKETOKEN",
+                        "Expiration": "2099-01-01T00:00:00Z",
+                        "AccountId": "01234567890",
+                    },
+                ),
+                patch(
+                    "botocore.credentials.ConfigProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.SharedCredentialProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.EnvProvider.load", return_value=None
+                ),
+            ],
+            'z',
+        ),
+        # Test case 3: Credentials set via Environment variables
+        (
+            {
+                'AWS_ACCESS_KEY_ID': 'FAKEACCESSKEY',
+                'AWS_SECRET_ACCESS_KEY': 'FAKESECRET',
+                'AWS_SESSION_TOKEN': 'FAKETOKEN',
+            },
+            {},
+            [],
+            'g',
+        ),
+        # Test case 4: Credentials set via code
+        (
+            {},
+            {
+                'aws_access_key_id': 'FAKEACCESSKEY',
+                'aws_secret_access_key': 'FAKESECRET',
+                'aws_session_token': 'FAKETOKEN',
+            },
+            [],
+            'e',
+        ),
+        # Test case 5: Credentials set with Boto2 config
+        (
+            {},
+            {},
+            [
+                patch(
+                    "botocore.configloader.raw_config_parse",
+                    return_value={
+                        "Credentials": {
+                            "aws_access_key_id": "FAKEACCESSKEY",
+                            "aws_secret_access_key": "FAKESECRETKEY",
+                        }
+                    },
+                ),
+                patch(
+                    "botocore.credentials.ConfigProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.SharedCredentialProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.EnvProvider.load", return_value=None
+                ),
+            ],
+            'x',
+        ),
+        # Test case 6: Credentials set via process provider
+        (
+            {},
+            {},
+            [
+                patch(
+                    "botocore.credentials.ProcessProvider._retrieve_credentials_using",
+                    return_value={
+                        'access_key': "FAKEACCESSKEY",
+                        'secret_key': "FAKESECRETKEY",
+                        'token': "FAKETOKEN",
+                    },
+                ),
+                patch(
+                    "botocore.credentials.ProcessProvider._credential_process",
+                    return_value="Mock_credential_process",
+                ),
+                patch(
+                    "botocore.credentials.ContainerProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.ConfigProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.SharedCredentialProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.EnvProvider.load", return_value=None
+                ),
+            ],
+            ['v', 'w'],
+        ),
+    ],
+)
+def test_user_agent_feature_ids(
+    environ_vars,
+    fake_credentials,
+    patches,
+    expected_feature_id,
+    monkeypatch,
+    patched_session,
+):
+    for var, value in environ_vars.items():
+        monkeypatch.setenv(var, value)
+
+    for patch_obj in patches:
+        patch_obj.start()
+
+    try:
+        client = patched_session.create_client(
+            "s3", region_name="us-east-1", **fake_credentials
+        )
+        _assert_feature_ids_in_ua(client, expected_feature_id)
+    finally:
+        for patch_obj in patches:
+            patch_obj.stop()
+
+
+@pytest.mark.parametrize(
+    "creds_env_var,creds_file_content,patches,expected_feature_id",
+    [
+        (
+            'AWS_SHARED_CREDENTIALS_FILE',
+            '[default]\naws_access_key_id = FAKEACCESSKEY\naws_secret_access_key = FAKESECRET',
+            [
+                patch(
+                    "botocore.credentials.AssumeRoleProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.EnvProvider.load", return_value=None
+                ),
+            ],
+            'n',
+        ),
+        (
+            'AWS_CONFIG_FILE',
+            '[default]\naws_access_key_id = FAKEACCESSKEY\naws_secret_access_key = FAKESECRET',
+            [
+                patch(
+                    "botocore.credentials.AssumeRoleProvider.load",
+                    return_value=None,
+                ),
+                patch(
+                    "botocore.credentials.EnvProvider.load", return_value=None
+                ),
+                patch(
+                    "botocore.credentials.SharedCredentialProvider.load",
+                    return_value=None,
+                ),
+            ],
+            'n',
+        ),
+    ],
+)
+def test_user_agent_has_file_based_feature_ids(
+    creds_env_var,
+    creds_file_content,
+    patches,
+    expected_feature_id,
+    tmp_path,
+    monkeypatch,
+):
+    credentials_file = tmp_path / "creds"
+    credentials_file.write_text(creds_file_content)
+    monkeypatch.setenv(creds_env_var, str(credentials_file))
+
+    for patch_obj in patches:
+        patch_obj.start()
+
+    try:
+        session = Session()
+        client = session.create_client("s3", region_name="us-east-1")
+        _assert_feature_ids_in_ua(client, expected_feature_id)
+    finally:
+        for patch_obj in patches:
+            patch_obj.stop()
+
+
+def _assert_feature_ids_in_ua(client, expected_feature_ids):
+    """Helper to test feature IDs appear in user agent for multiple calls."""
+    with ClientHTTPStubber(client, strict=True) as http_stubber:
+        http_stubber.add_response()
+        http_stubber.add_response()
+        client.list_buckets()
+        client.list_buckets()
+
+    ua_strings = get_captured_ua_strings(http_stubber)
+    for ua_string in ua_strings:
+        feature_list = parse_registered_feature_ids(ua_string)
+        for expected_id in expected_feature_ids:
+            assert expected_id in feature_list
+
+
+@pytest.mark.parametrize(
+    "config_content,env_vars,expected_source_features,expected_provider_feature",
+    [
+        # Test Case 1: Assume Role with source profile
+        (
+            '''[profile assume-role-test]
+role_arn = arn:aws:iam::123456789012:role/test-role
+source_profile = base
+
+[profile base]
+aws_access_key_id = FAKEACCESSKEY
+aws_secret_access_key = FAKESECRET''',
+            {},
+            [
+                'n',  # CREDENTIALS_PROFILE
+                'o',  # CREDENTIALS_PROFILE_SOURCE_PROFILE
+            ],
+            'i',  # CREDENTIALS_STS_ASSUME_ROLE
+        ),
+        # Test Case 2: Assume Role with named provider
+        (
+            '''[profile assume-role-test]
+role_arn = arn:aws:iam::123456789012:role/test-role
+credential_source = Environment''',
+            {
+                'AWS_ACCESS_KEY_ID': 'FAKEACCESSKEY',
+                'AWS_SECRET_ACCESS_KEY': 'FAKESECRET',
+            },
+            [
+                'g',  # CREDENTIALS_ENV_VARS
+                'p',  # CREDENTIALS_PROFILE_NAMED_PROVIDER
+            ],
+            'i',  # CREDENTIALS_STS_ASSUME_ROLE
+        ),
+    ],
+)
+def test_user_agent_has_assume_role_feature_ids(
+    config_content,
+    env_vars,
+    expected_source_features,
+    expected_provider_feature,
+    tmp_path,
+):
+    session = _create_assume_role_session(config_content, tmp_path)
+
+    # Set env vars if needed
+    with patch.dict(os.environ, env_vars, clear=True):
+        with SessionHTTPStubber(session) as stubber:
+            s3 = session.create_client('s3', region_name='us-east-1')
+            _add_assume_role_http_response(stubber, with_web_identity=False)
+            stubber.add_response()
+            stubber.add_response()
+            s3.list_buckets()
+            s3.list_buckets()
+
+    ua_strings = get_captured_ua_strings(stubber)
+    _assert_deferred_credential_feature_ids(
+        ua_strings, expected_source_features, expected_provider_feature
     )
-    @patch("botocore.credentials.ContainerProvider.load", return_value=None)
-    @patch("botocore.credentials.ConfigProvider.load", return_value=None)
-    @patch(
-        "botocore.credentials.SharedCredentialProvider.load", return_value=None
+
+
+@pytest.mark.parametrize(
+    "config_content,env_vars,expected_source_features,expected_provider_feature",
+    [
+        # Test Case 1: Assume Role with Web Identity through config profile
+        (
+            '''[profile assume-role-test]
+role_arn = arn:aws:iam::123456789012:role/test-role
+web_identity_token_file = {token_file}''',
+            {},
+            ['q'],  # CREDENTIALS_PROFILE_STS_WEB_ID_TOKEN
+            'k',  # CREDENTIALS_STS_ASSUME_ROLE_WEB_ID
+        ),
+        # Test Case 2: Assume Role with Web Identity through env vars
+        (
+            '',
+            {
+                'AWS_ROLE_ARN': 'arn:aws:iam::123456789012:role/test-role',
+                'AWS_WEB_IDENTITY_TOKEN_FILE': '{token_file}',
+                'AWS_ROLE_SESSION_NAME': 'test-session',
+            },
+            ['h'],  # CREDENTIALS_ENV_VARS_STS_WEB_ID_TOKEN
+            'k',  # CREDENTIALS_STS_ASSUME_ROLE_WEB_ID
+        ),
+    ],
+)
+def test_user_agent_has_assume_role_with_web_identity_feature_ids(
+    config_content,
+    env_vars,
+    expected_source_features,
+    expected_provider_feature,
+    tmp_path,
+):
+    token_file = tmp_path / 'token.jwt'
+    token_file.write_text('fake-jwt-token')
+    if 'AWS_WEB_IDENTITY_TOKEN_FILE' in env_vars:
+        env_vars['AWS_WEB_IDENTITY_TOKEN_FILE'] = str(token_file)
+    elif config_content and 'web_identity_token_file' in config_content:
+        config_content = config_content.replace(
+            '{token_file}', str(token_file)
+        )
+
+    session = _create_assume_role_session(config_content, tmp_path)
+
+    # Set env vars if needed
+    with patch.dict(os.environ, env_vars, clear=True):
+        with SessionHTTPStubber(session) as stubber:
+            s3 = session.create_client('s3', region_name='us-east-1')
+            _add_assume_role_http_response(stubber, with_web_identity=True)
+            stubber.add_response()
+            stubber.add_response()
+            s3.list_buckets()
+            s3.list_buckets()
+
+    ua_strings = get_captured_ua_strings(stubber)
+    _assert_deferred_credential_feature_ids(
+        ua_strings, expected_source_features, expected_provider_feature
     )
-    @patch("botocore.credentials.EnvProvider.load", return_value=None)
-    def test_user_agent_has_imds_credentials_feature_id(
-        self,
-        _unused_mock_env_load,
-        _unused_mock_shared_load,
-        _unused_mock_config_load,
-        _unused_mock_container_load,
-        mock_retrieve_iam_role_credentials,
-        patched_session,
-    ):
-        fake_creds = {
-            "role_name": "FAKEROLE",
-            "access_key": "FAKEACCESSKEY",
-            "secret_key": "FAKESECRET",
-            "token": "FAKETOKEN",
-            "expiry_time": "2099-01-01T00:00:00Z",
-        }
-        mock_retrieve_iam_role_credentials.return_value = fake_creds
 
-        client = patched_session.create_client("s3", region_name="us-east-1")
-        with ClientHTTPStubber(client, strict=True) as http_stubber:
-            # We want to call this twice to assert that the feature id exists
-            # for multiple calls with the same credentials
-            http_stubber.add_response()
-            http_stubber.add_response()
-            client.list_buckets()
-            client.list_buckets()
 
-        ua_string = get_captured_ua_strings(http_stubber)
-        feature_list_one = parse_registered_feature_ids(ua_string[0])
-        feature_list_two = parse_registered_feature_ids(ua_string[1])
-        assert '0' in feature_list_one and '0' in feature_list_two
+def _create_assume_role_session(config_content, tmp_path):
+    if config_content:
+        config_file = tmp_path / 'config'
+        config_file.write_text(config_content)
+        session = Session(profile='assume-role-test')
+        session.set_config_variable('config_file', str(config_file))
+    else:
+        session = Session()
+    return session
 
-    @patch("botocore.credentials.ContainerMetadataFetcher.retrieve_full_uri")
-    @patch("botocore.credentials.ConfigProvider.load", return_value=None)
-    @patch(
-        "botocore.credentials.SharedCredentialProvider.load", return_value=None
+
+def _add_assume_role_http_response(stubber, with_web_identity):
+    """Add HTTP response for AssumeRole or AssumeRoleWithWebIdentity call with proper credentials"""
+    expiration = (datetime.now(timezone.utc) + timedelta(hours=1)).strftime(
+        '%Y-%m-%dT%H:%M:%SZ'
     )
-    @patch("botocore.credentials.EnvProvider.load", return_value=None)
-    def test_user_agent_has_http_credentials_feature_id(
-        self,
-        _unused_mock_env_load,
-        _unused_mock_shared_load,
-        _unused_mock_config_load,
-        mock_load_http_credentials,
-        monkeypatch,
-        patched_session,
-    ):
-        environ = {
-            'AWS_CONTAINER_CREDENTIALS_FULL_URI': 'http://localhost/foo',
-            'AWS_CONTAINER_AUTHORIZATION_TOKEN': 'Basic auth-token',
-        }
-        for var in environ:
-            monkeypatch.setenv(var, environ[var])
+    method_name = (
+        'AssumeRoleWithWebIdentity' if with_web_identity else 'AssumeRole'
+    )
+    body = (
+        f'<{method_name}Response>'
+        f'  <{method_name}Result>'
+        '    <AssumedRoleUser>'
+        '      <Arn>arn:aws:sts::123456789012:user</Arn>'
+        '      <AssumedRoleId>AKID:test-session-123</AssumedRoleId>'
+        '    </AssumedRoleUser>'
+        '    <Credentials>'
+        f'      <AccessKeyId>FAKEASSUMEROLEKEY</AccessKeyId>'
+        f'      <SecretAccessKey>FAKEASSUMEROLSECRET</SecretAccessKey>'
+        '      <SessionToken>FAKETOKEN</SessionToken>'
+        f'      <Expiration>{expiration}</Expiration>'
+        '    </Credentials>'
+        f'  </{method_name}Result>'
+        f'</{method_name}Response>'
+    )
+    stubber.add_response(body=body.encode('utf-8'))
 
-        fake_creds = {
-            "AccessKeyId": "FAKEACCESSKEY",
-            "SecretAccessKey": "FAKESECRET",
-            "Token": "FAKETOKEN",
-            "Expiration": "2099-01-01T00:00:00Z",
-            "AccountId": "01234567890",
-        }
-        mock_load_http_credentials.return_value = fake_creds
 
-        client = patched_session.create_client("s3", region_name="us-east-1")
-        with ClientHTTPStubber(client, strict=True) as http_stubber:
-            # We want to call this twice to assert that the feature id exists
-            # for multiple calls with the same credentials
-            http_stubber.add_response()
-            http_stubber.add_response()
-            client.list_buckets()
-            client.list_buckets()
+def _assert_deferred_credential_feature_ids(
+    ua_strings,
+    expected_source_features,
+    expected_provider_feature,
+):
+    """Helper to assert feature IDs for deferred credential provider tests"""
+    assert len(ua_strings) == 3
 
-        ua_string = get_captured_ua_strings(http_stubber)
-        feature_list_one = parse_registered_feature_ids(ua_string[0])
-        feature_list_two = parse_registered_feature_ids(ua_string[1])
-        assert 'z' in feature_list_one and 'z' in feature_list_two
+    # Request to fetch credentials should only register feature ids for the credential source
+    credential_source_feature_list = parse_registered_feature_ids(
+        ua_strings[0]
+    )
+    for feature in expected_source_features:
+        assert feature in credential_source_feature_list
+    assert expected_provider_feature not in credential_source_feature_list
+
+    # Original operation request should register feature ids for both the credential source and the provider
+    for i in [1, 2]:
+        operation_feature_list = parse_registered_feature_ids(ua_strings[i])
+        for feature in expected_source_features:
+            assert feature in operation_feature_list
+        assert expected_provider_feature in operation_feature_list
